import { makeEncode, makeRecipes, readMsg, decode, decodeWithOptions, } from "../commands/index.js";
import { knownUSPVersions, makeCallbackRouter, makeRouter, parseCallbackOptions, } from "../util.js";
const defaultPublishEndpoint = "/usp/endpoint";
const defaultSubscribeEndpoint = "/usp/controller";
const defaultIdEndpoint = "obuspa/EndpointID";
const defaultFromId = "proto::interop-usp-controller";
const defaultConnectionTimeout = 5000;
const defaultIdResolveTimeout = 5000;
const fixId = (s) => s.split("+").join("%2B");
const wait = (ms) => {
let waitId;
let waitPromise = new Promise((_, reject) => {
waitId = setTimeout(() => {
reject(`timeout of ${ms}ms reached`);
}, ms);
});
return [waitId, waitPromise];
};
const timeoutWrapper = (promise, ms) => {
if (typeof ms === "number" && ms >= 0) {
const [waitId, waitPromise] = wait(ms);
return Promise.race([
promise().then((res) => {
clearTimeout(waitId);
return res;
}),
waitPromise,
]);
}
else
return promise();
};
const applyCommandChanges = (opts, cmdName, cmd, args) => {
if (opts.get?.retainPath === true && cmdName === "get") {
const newOptions = args.length > 1
? "retainPath" in args[1]
? args[1]
: { retainPath: true, ...args[1] }
: {};
return cmd(...[args[0], newOptions, ...args.slice(2)]);
}
return cmd(...args);
};
const wrap = (opts, cmdName, cmd) => {
return ((...args) => {
// Promise.resolve added to handle non-promise commands
const cmdCall = () => Promise.resolve(applyCommandChanges(opts, cmdName, cmd, args));
const finalCall = () => timeoutWrapper(cmdCall, opts.timeout)
.then((res) => {
opts.postCall && opts.postCall(cmdName, args, res);
return res;
})
.catch((err) => {
opts.postCall && opts.postCall(cmdName, args, err);
throw err;
});
if (opts.preCall)
opts.preCall(cmdName, args);
return finalCall();
});
};
const addOptions = (usp, opts) => {
const mainUSP = Object.entries(usp).reduce((acc, [k, fn]) => ({
...acc,
[k]: wrap(opts, k, fn),
}), {});
mainUSP.options = (opts) => addOptions(mainUSP, opts);
return mainUSP;
};
const correctOptions = (options) => ({
...options,
reconnectPeriod: options.closeOnDisconnect ? 0 : 1000,
});
const buildConnect = ({ connectClient, decodeID, loadProtobuf }) => async (options, events) => {
const subscribeEndpoint = options.subscribeEndpoint || defaultSubscribeEndpoint;
const publishEndpoint = options.publishEndpoint || defaultPublishEndpoint;
const idEndpoint = options.idEndpoint || defaultIdEndpoint;
const connectTimeout = options.connectTimeout || defaultConnectionTimeout;
const correctedOptions = correctOptions(options);
const client = await timeoutWrapper(() => connectClient(correctedOptions), connectTimeout);
let version = options.version || "1.1";
// Note !== false is used since useLatestUSPVersion is assumed to be true
//  needs version check only if no version was provided or useLatestUSPVersion is true
const needsVersionCheck = typeof options.version !== "string" || options.useLatestUSPVersion;
let proto = await loadProtobuf(version);
const router = makeRouter();
const callbackRouter = makeCallbackRouter();
const handleError = (err) => events && events.onError && events.onError(err);
const handleOffline = () => {
events && events.onOffline && events.onOffline();
client.end();
};
const handleReconnect = () => events && events.onReconnect && events.onReconnect();
const handleClose = () => events && events.onClose && events.onClose();
callbackRouter.add("error", handleError);
const handleInit = () => new Promise((resolve, reject) => {
const id = setTimeout(() => reject({
errMsg: `toId was not received within timeout(${defaultIdResolveTimeout}ms)`,
}), defaultIdResolveTimeout);
client.on("message", (topic, data) => {
clearTimeout(id);
client.unsubscribe(idEndpoint);
resolve(decodeID(data || topic));
});
client.subscribe(idEndpoint);
});
const toId = fixId(options.toId || (await handleInit()));
const fromId = options.fromId || defaultFromId;
client.on("message", (_topic, data) => {
const parsedMsg = readMsg(proto, data || _topic);
const [id, message, err, cmdType] = decode(parsedMsg, version);
if (typeof id !== "string") {
handleError(`Could not locate id for message:\n${JSON.stringify(parsedMsg, null, 2)}`);
}
else {
const call = router.get(id, cmdType || "NOTIFY");
if (call && call.resolve && call.reject) {
if (err)
call.reject(err);
else if (call.options) {
const messageAfterOptions = decodeWithOptions(parsedMsg, cmdType, call.options, version);
call.resolve(messageAfterOptions);
}
else
call.resolve(message);
}
const cbs = callbackRouter.get(id);
const callbackOptions = parseCallbackOptions(parsedMsg);
cbs.forEach((cb) => {
if (callbackOptions)
cb(message || err || "", parsedMsg, callbackOptions);
else
cb(message || err || "", parsedMsg);
});
}
});
client.on("error", (err) => {
callbackRouter.get("error").forEach((cb) => cb(err));
handleError(JSON.stringify(err, null, 2));
});
client.on("offline", handleOffline);
client.on("reconnect", () => {
handleReconnect();
});
client.on("close", () => {
handleClose();
});
const on = (ident, callback) => {
callbackRouter.add(ident, callback);
return () => {
callbackRouter.del(ident);
};
};
const encode = makeEncode(proto, { fromId, toId });
const call = (command, args, callOpts) => new Promise((resolve, reject) => {
const [id, msg, err] = encode(command, version, args);
if (err)
reject(err);
else {
router.add(id, callOpts?.responseMsgType || command, {
resolve,
reject,
options: args.options,
});
client.publish(publishEndpoint, msg);
}
});
await client.subscribe(subscribeEndpoint);
const baseUSP = {
get: (paths, options) => call("GET", { paths, options }),
getNested: (paths, options) => call("GET", { paths, options: { ...options, retainPath: true } }),
set: (path, value, options) => call("SET", { path, value, options }),
add: (path, value, options) => call("ADD", { path, value, options }),
del: (paths, allowPartial) => call("DELETE", { paths, allowPartial }),
instances: (paths, opts) => call("GET_INSTANCES", { paths, opts }),
supportedDM: (paths, opts) => call("GET_SUPPORTED_DM", { paths, opts }),
supportedProto: (proto) => call("GET_SUPPORTED_PROTO", { proto }),
_operate: (path, id, resp, input, commandKey) => call("OPERATE", { path, input, id, resp, commandKey }),
_notifyResp: (subscriptionId) => call("NOTIFY_RESP", { subscriptionId }),
on,
...makeRecipes(call, on),
disconnect: () => client.end(),
getUSPVersion: () => version,
setUSPVersion: (v) => {
version = v;
},
};
if (needsVersionCheck) {
const versions = await baseUSP.supportedProto();
const latestKnownVersion = versions
.reverse()
.find((v) => knownUSPVersions.includes(v));
if (latestKnownVersion) {
proto = await loadProtobuf(latestKnownVersion);
version = latestKnownVersion;
}
}
return {
...baseUSP,
options: (opts) => addOptions(baseUSP, opts),
};
};
export default buildConnect;
